/**
 * @file
 *
 * @ingroup arm_beagle
 *
 * @brief Startup code.
 */

/*
 * Copyright (c) 2014 Ben Gras <beng@shrike-systems.com>. All rights reserved.
 *
 * The license and distribution terms for this file may be
 * found in the file LICENSE in this distribution or at
 * http://www.rtems.com/license/LICENSE.
 */

#include <bsp/bootcard.h>
#include <bsp.h>
#include <bsp/irq-generic.h>
#include <bsp/irq.h>
#include <bsp/linker-symbols.h>
#include <bspopts.h>
#include <bsp/start.h>
#include <bsp/uart-output-char.h>
#include <libchip/serial.h>
#include <libcpu/arm-cp15.h>
#include <stdbool.h>

#define AM335X_CM_BASE 0x44E00000
#define AM335X_CM_SIZE 0x1000

#define AM335X_PRM_DEVICE_OFFSET 0xf00
#define AM335X_PRM_RSTCTRL_REG 0x00
#define AM335X_RST_GLOBAL_WARM_SW_BIT 0

#define DM37XX_CM_BASE 0x48307000
#define DM37XX_CM_SIZE 0x1000
#define DM37XX_PRM_RSTCTRL_REG 0x250
#define DM37XX_RST_DPLL3_BIT 2

BSP_START_TEXT_SECTION void bsp_start_hook_0(void)
{
}

/* From shared ARM BSP startup code */
extern uint32_t bsp_arm_start_initial_r0, bsp_arm_start_initial_r1;

BSP_START_TEXT_SECTION void bsp_start_hook_1()
{
  bsp_start_clear_bss();
  BSP_CONSOLE_UART_INIT(UART_CLOCK / 16 / CONSOLE_BAUD);
}

#define ARM_SECTIONS		4096	/* all sections needed to describe the virtual address space */
#define ARM_SECTION_SIZE        (1024 * 1024)	/* how much virtual memory is described by one section */

static void pg_set(uint32_t *pg, uint32_t start, uint32_t end, uint32_t flags)
{
  int i;

  for(i = start/ARM_SECTION_SIZE; i < end/ARM_SECTION_SIZE; i++) {
    uint32_t mapaddr = i * ARM_SECTION_SIZE;

    if(flags) pg[i] = mapaddr | flags;
    else pg[i] = 0;
  }
}

extern uint32_t RamBase, RamSize;	/* from linker script */

static  void bsp_caching_and_paging(void)
{
        uint32_t sctlr;
        uint32_t actlr;
  int i;
  static uint32_t pagetable[ARM_SECTIONS] __attribute__((aligned (1024*16)));
  uint32_t rambase = (uint32_t) &RamBase, ramsize = (uint32_t) &RamSize;
  uint32_t *c;

  /* Set up pagetable. It is going to be:
   *   - empty
   * except:
   *   - 1:1 cacheable mapping for physical RAM
   *   - 1:1 uncacheable mapping for device memory
   *
   * this way:
   *   - caches can be on, devices can be accessed
   *   - invalid RAM accesses are trapped
   */

  /* clear all */
  pg_set(pagetable, 0, 0xFFFFF000, 0);

  /* map RAM cacheable */
  pg_set(pagetable, rambase, rambase+ramsize,
    ARM_VM_SECTION_PRESENT | ARM_VM_SECTION_USER | ARM_VM_SECTION_CACHED);

  /* map device memory uncacheable */
  pg_set(pagetable, BSP_DEVICEMEM_START, BSP_DEVICEMEM_END,
  ARM_VM_SECTION_PRESENT | ARM_VM_SECTION_USER | ARM_VM_SECTION_DEVICE);

  /* load our pagetable */
  write_ttbr0((uint32_t) pagetable);

        /* Set all Domains to Client */
        write_dacr(0x55555555);

        sctlr = read_sctlr();

        /* Enable MMU */
        sctlr |= CPU_CONTROL_MMU_ENABLE;

        /* TRE set to zero (default reset value): TEX[2:0] are used, plus C and B bits.*/
        sctlr &= ~CPU_CONTROL_TR_ENABLE;

        /* AFE set to zero (default reset value): not using simplified model. */
        sctlr &= ~CPU_CONTROL_AF_ENABLE;

        /* Enable instruction, data cache and branch prediction */
        sctlr |= CPU_CONTROL_DC_ENABLE;
        sctlr |= CPU_CONTROL_IC_ENABLE;
        sctlr |= CPU_CONTROL_BPRD_ENABLE;

        /* Enable barriers */
        sctlr |= CPU_CONTROL_32BD_ENABLE;

        /* Enable L2 cache (cortex-a8) */
#define CORTEX_A8_L2EN   (0x02)
        actlr = read_actlr();
        actlr |= CORTEX_A8_L2EN;

        write_actlr(actlr);

  /* set 16kB TTBR alignment */
  write_ttbcr(0);

        write_sctlr(sctlr);


  /* set 16kB TTBR alignment */
  write_ttbcr(0);


  /* install our pagetable, also clears TLB */
  write_ttbr0((uint32_t) pagetable);
}

void bsp_start(void)
{
//  bsp_caching_and_paging();
  bsp_interrupt_initialize();
printk("c&p %s:%d\n", __FILE__, __LINE__);
}

void
bsp_reset(void)
{
#if IS_OMAP3
  static uint32_t reset_base = DM37XX_CM_BASE;
  mmio_set((reset_base + DM37XX_PRM_RSTCTRL_REG),
      (1 << DM37XX_RST_DPLL3_BIT));
#endif

#if IS_AM335X
  static uint32_t reset_base = AM335X_CM_BASE;
  mmio_set((reset_base + AM335X_PRM_DEVICE_OFFSET +
    AM335X_PRM_RSTCTRL_REG),
      (1 << AM335X_RST_GLOBAL_WARM_SW_BIT));
#endif
}

